#ifndef __ASM_CRIS_PCI_H
#define __ASM_CRIS_PCI_H

/* ETRAX chips don't have a PCI bus. This file is just here because some stupid .c code
 * includes it even if CONFIG_PCI is not set.
 */

#endif /* __ASM_CRIS_PCI_H */

