/*
 * include/sgidefs.h
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1996 by Ralf Baechle
 *
 * This file is here to satisfy GCC's expectations.
 */
#ifndef __SGIDEFS_H
#define __SGIDEFS_H

#include <asm/sgidefs.h>

#endif /* __SGIDEFS_H */
