/*
 *  linux/include/asm-arm/arch-ebsa110/irqs.h
 *
 *  Copyright (C) 1996 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#define NR_IRQS			8

#define IRQ_EBSA110_PRINTER	0
#define IRQ_EBSA110_COM1	1
#define IRQ_EBSA110_COM2	2
#define IRQ_EBSA110_ETHERNET	3
#define IRQ_EBSA110_TIMER0	4
#define IRQ_EBSA110_TIMER1	5
#define IRQ_EBSA110_PCMCIA	6
#define IRQ_EBSA110_IMMEDIATE	7
