/* $Id: isdn_ttyfax.h,v 1.2.6.1 2001/09/23 22:24:32 kai Exp $
 *
 * header for Linux ISDN subsystem, tty_fax related functions (linklevel).
 *
 * Copyright 1999   by Armin Schindler (mac@melware.de)
 * Copyright 1999   by Ralf Spachmann (mel@melware.de)
 * Copyright 1999   by Cytronics & Melware
 *
 * This software may be used and distributed according to the terms
 * of the GNU General Public License, incorporated herein by reference.
 *
 */


#define XON	0x11
#define XOFF	0x13
#define DC2	0x12

