
/* this file was generated on Thu Jan  8 00:21:21 GMT 1998  */


/* size of transition table is 898 bytes */

static char timertr_entry [ ] = {
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , ADM           ,
   NOP           , ADM           , NOP           , ADM           , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , CONN          , NOP           , CONN          ,
   NOP           , CONN          , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_WAIT    , NOP           , RESET_WAIT    , NOP           , RESET_WAIT    ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , NOP           , RESET_CHECK   ,
   NOP           , RESET_CHECK   , NOP           , RESET_CHECK   , SETUP7        , SETUP         ,
   SETUP7        , SETUP         , SETUP3        , NORMAL        , SETUP3        , NORMAL        ,
   SETUP8        , ADM           , SETUP8        , ADM           , SETUP3        , NORMAL        ,
   SETUP3        , NORMAL        , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , NOP           , SETUP         , NOP           , SETUP         ,
   NOP           , SETUP         , RESET7        , RESET         , RESET7        , RESET         ,
   RESET3        , NORMAL        , RESET3        , NORMAL        , RESET8        , ADM           ,
   RESET8        , ADM           , RESET3        , NORMAL        , RESET3        , NORMAL        ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   NOP           , RESET         , NOP           , RESET         , NOP           , RESET         ,
   D_CONN6       , D_CONN        , D_CONN6       , D_CONN        , D_CONN6       , D_CONN        ,
   D_CONN6       , D_CONN        , D_CONN7       , ADM           , D_CONN7       , ADM           ,
   D_CONN7       , ADM           , D_CONN7       , ADM           , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , NOP           , D_CONN        ,
   NOP           , D_CONN        , NOP           , D_CONN        , ERR7          , ERROR         ,
   ERR7          , ERROR         , ERR7          , ERROR         , ERR7          , ERROR         ,
   ERR8          , RESET_WAIT    , ERR8          , RESET_WAIT    , ERR8          , RESET_WAIT    ,
   ERR8          , RESET_WAIT    , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NOP           , ERROR         , NOP           , ERROR         ,
   NOP           , ERROR         , NORMAL20      , AWAIT         , NOP           , NORMAL        ,
   NORMAL20      , AWAIT         , NOP           , NORMAL        , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   NORMAL19      , NORMAL        , NORMAL19      , NORMAL        , NORMAL19      , NORMAL        ,
   NORMAL19      , NORMAL        , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , NOP           , NORMAL        ,
   NOP           , NORMAL        , NOP           , NORMAL        , NOP           , NORMAL        ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , NORMAL20      , AWAIT         , NOP           , NORMAL        ,
   NORMAL20      , AWAIT         , NOP           , NORMAL        , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   BUSY24        , AWAIT_BUSY    , NOP           , BUSY          , BUSY24        , AWAIT_BUSY    ,
   NOP           , BUSY          , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , BUSY23        , BUSY          ,
   BUSY23        , BUSY          , BUSY23        , BUSY          , BUSY23        , BUSY          ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , BUSY25        , BUSY          , BUSY26        , BUSY          ,
   BUSY25        , BUSY          , BUSY26        , BUSY          , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   NOP           , BUSY          , NOP           , BUSY          , NOP           , BUSY          ,
   NOP           , BUSY          , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , NOP           , REJECT        ,
   NOP           , REJECT        , NOP           , REJECT        , NOP           , REJECT        ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , NOP           , REJECT        , NOP           , REJECT        ,
   NOP           , REJECT        , NOP           , REJECT        , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   NOP           , REJECT        , NOP           , REJECT        , NOP           , REJECT        ,
   NOP           , REJECT        , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , NOP           , REJECT        ,
   NOP           , REJECT        , NOP           , REJECT        , NOP           , REJECT        ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , NOP           , AWAIT         , NOP           , AWAIT         ,
   NOP           , AWAIT         , NOP           , AWAIT         , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   NOP           , AWAIT         , NOP           , AWAIT         , NOP           , AWAIT         ,
   NOP           , AWAIT         , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , NOP           , AWAIT         ,
   NOP           , AWAIT         , NOP           , AWAIT         , NOP           , AWAIT         ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , NOP           , AWAIT         , NOP           , AWAIT         ,
   NOP           , AWAIT         , NOP           , AWAIT         , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    ,
   NOP           , AWAIT_BUSY    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , NOP           , AWAIT_BUSY    ,
   NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    ,
   NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    , NOP           , AWAIT_BUSY    ,
   NOP           , AWAIT_BUSY    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , NOP           , AWAIT_REJECT  ,
   NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  ,
   NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  ,
   NOP           , AWAIT_REJECT  , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , NOP           , AWAIT_REJECT  ,
   NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  , NOP           , AWAIT_REJECT  ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT    , SH11          , RESET_WAIT    ,
   SH11          , RESET_WAIT    , SH11          , RESET_WAIT     };
