/*
 *  linux/arch/arm/boot/compressed/head-netwinder.S
 *
 *  Copyright (C) 2000-2002 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
		.section	".start", "ax"

		mov	r7, #5
		mov	r8, #0
